import genj.gedcom.*;
import genj.report.*;
import java.io.*;

/**
 * GenJ - Report
 * MigrateToMedias
 * @author Nils Meier nils@meiers.org
 * @version 0.1
 */
public class ReportMigrateToMedias implements Report {

  private static final String tagsToMigrate[] = { "PHOT" };

  /**
   * Returns the name of this report - should be localized.
   */
  public String getName() {
    return "Migrate to Medias";
  }

  /**
   * Some information about this report
   * @return Information as String
   */
  public String getInfo() {
    return "This report migrates an existing Gedcom file to use "+
           "multimedia properties for referencing external files "+
           "like pictures. Currently only property PHOT is substituted "+
           "by the corresponding OBJE|FILE solution";
  }

  /**
   * This method actually starts this report
   */
  public boolean start(ReportView view, Gedcom gedcom, ReportWriter out) throws InterruptedException {

    // Looking for individuals and families
    out.println("***Looking for individuals");
    EntityList indis = gedcom.getEntities(Gedcom.INDIVIDUALS);
    if (indis.getSize()==0)
      out.println("- none found");
    else {
      for (int i=0;i<indis.getSize();i++)
        analyseEntity(indis.getIndi(i),out);
    }
    out.println();

    out.println("***Looking for families");
    EntityList fams = gedcom.getEntities(Gedcom.FAMILIES);
    if (fams.getSize()==0)
      out.println("- none found");
    else {
      for (int i=0;i<fams.getSize();i++)
        analyseEntity(fams.getFam(i),out);
    }
    out.println();

    out.println("***Done");

    // Done
    return true;
  }

  /**
   * Helper that scans entity for properties to migrate
   */
  private void analyseEntity(Entity entity, ReportWriter out) throws InterruptedException {

    // Get properties of entity
    if (entity.getProperty().getProperties().getSize()==0)
      return;

    // Look for properties to migrate
    PropertyList props = entity.getProperty().getProperties();
    for (int i=props.getSize()-1;i>=0;i--)
      analyseProperty(props.get(i),out);

    // Done
  }

  /**
   * Helper that checks property as candidate to migrate
   */
  private void analyseProperty(Property prop, ReportWriter out)  throws InterruptedException {

    // Compare to tags to migrate
    for (int i=0;i<tagsToMigrate.length;i++) {
      if (prop.getTag().equals(tagsToMigrate[i]))
        migrateProperty(prop,out);
    }

    // Done
  }

  /**
   * Helper that migrates property
   */
  private void migrateProperty(Property prop, ReportWriter out) throws InterruptedException {

    // Properties with sub-properties are not migrated
    if (prop.getNoOfProperties()>0) {
      out.println("Skipping "+prop.getTag()+" of entity "+prop.getEntity().getId());
      return;
    }

    // Migrate it
    String value = prop.getValue();
    out.println("Migrating "+prop.getTag()+" of entity "+prop.getEntity().getId()+" pointing to "+value);

    // .. delete old
    Property parent = prop.getParent();
    parent.delProperty(prop);

    // .. add as new OBJE|FILE
    Property pobje = new PropertyMedia(null);
    pobje .addProperty(new PropertyFile(value));
    parent.addProperty(pobje);

    // Done
  }

  /**
   * Indication of how this reports shows information
   * to the user. Standard Out here only.
   */
  public boolean usesStandardOut() {
    return true;
  }

  /**
   * Author
   */
  public String getAuthor() {
    return "Nils Meier <nils@meiers.org>";
  }

  /**
   * Tells wether this report doesn't change information in the Gedcom-file
   */
  public boolean isReadOnly() {
    return false;
  }

}

