package genj.edit;

import javax.swing.*; 
import javax.swing.tree.*; 
import javax.swing.event.*; 
import javax.swing.border.EmptyBorder; 
 
import java.awt.*;
import java.awt.event.*;
import java.util.*;

import genj.gedcom.*;
import genj.option.*; 
import genj.util.*;

/** 
 * Component for editing genealogic entity properties
 */
public class EditView extends JPanel implements ActionListener, TreeSelectionListener, GedcomListener {

  Gedcom    gedcom;
  Entity    entity;
  Frame     frame;
  
  private JPanel            actionPanel;
  private JButton           actionButtonAdd,
							actionButtonRemove,
							actionButtonUp, 
							actionButtonDown,
							actionButtonReturn;
  private JCheckBox         actionCheckStick;
  private JSplitPane        split;
	
  private JTree                  treeOfProps = null;
  private JScrollPane            paneForTree = null;
  private JPanel                 panelForProxy;
  
  private Proxy                  currentProxy = null;
  private Property               currentNode = null;

  private Vector            returnStack = new Vector(MAX_RETURN);
  private final static int  MAX_RETURN  = 10;
  
  private boolean keepSimple     = false;

  private Registry registry;

  private static ImageIcon imgStickOn,imgStickOff,imgReturn;

  static final Resources resources = new Resources("genj.edit");
  
  /**
   * Class for rendering tree cell nodes
   */
  class PropertyCellRenderer extends Component implements TreeCellRenderer {

	private boolean selected = false, focus = false;
	private String  tag, value;
	private ImgIcon image;
	private int GAP = 2;
	private Font    font;
	private Color   color;

	PropertyCellRenderer() {
	  JLabel label = new JLabel();
	  color=label.getForeground();
	  font =label.getFont      ();
	}

	/**
	  * This is messaged from JTree whenever it needs to get the size
	  * of the component or it wants to draw it.
	  * This attempts to set the font based on value, which will be
	  * a TreeNode.
	  */
	public Component getTreeCellRendererComponent(JTree tree, Object value,
												  boolean selected, boolean expanded,
												  boolean leaf, int row,
												  boolean hasFocus) {
  
	  // Set the text
	  Property prop = (Property)value;
	  tag = prop.getTag();
	  if (prop instanceof Entity)
		tag = "@"+((Entity)prop).getId()+"@ "+tag;

	  this.value = (prop.getValue()==null ? "" : prop.getValue());

	  // Set the image
	  image = prop.getImage(true);

	  // Done 
	  this.selected = selected;
	  this.focus = focus;

	  return this;
	}

	/**
	 * Returns the preferred size of this component
	 */
	public Dimension getPreferredSize() {

	  // Calculate Params
	  Graphics g = treeOfProps.getGraphics();
	  if (g==null)
		return new Dimension(0,0);

	  // Icon Width
	  int w = image.getIconWidth();

	  // Tag Width
	  g.setFont(new Font(font.getName(),font.BOLD,font.getSize()));
	  FontMetrics fm = g.getFontMetrics();
	  w += fm.stringWidth(tag);

	  // Value Width
	  g.setFont(new Font(font.getName(),font.PLAIN,font.getSize()));
	  fm = g.getFontMetrics();
	  w += fm.stringWidth(value);

	  // Height
	  int h = Math.max( image.getIconHeight() , fm.getHeight() );

	  // Done
	  return new Dimension(w+2*GAP,h);
	}

	/**
	  * Paint is subclassed to draw this treecell
	  */
	public void paint(Graphics g) {

	  // Color
	  Color            bColor;
	  if (selected)
		bColor = Color.yellow;
	  else
		bColor = treeOfProps.getBackground();
	  g.setColor(bColor);

	  // Parms
	  Dimension size = this.getSize();
	  int h = size.height,
		  w = size.width;

	  // Background
	  g.fillRect(0 , 0, w-1, h-1);

	  // Image
	  int iw = image.getIconWidth(),
		  ih = image.getIconHeight(),
		  ix = 0,
		  iy = (h-ih)/2;
	  image.paintIcon(g, ix, iy);

	  // Text
	  g.setColor(color);
	  g.setFont(new Font(font.getName(),font.BOLD,font.getSize()));
	  FontMetrics fm = g.getFontMetrics();
	  int tx = iw + GAP,
		  ty = (h+fm.getHeight())/2 - fm.getDescent();
	  g.drawString(tag, tx, ty);
	  tx += fm.stringWidth(tag) + GAP;
	  g.setFont(new Font(font.getName(),font.PLAIN,font.getSize()));
	  g.drawString(value, tx, ty);

	  // Done
	}
  }

/**
 * GenEdit constructor comment.
 */
public EditView(Gedcom setGedcom, Registry setRegistry, Frame setFrame) {

	// Super constructor
	super();

	this.gedcom   = setGedcom;
	this.frame    = setFrame;
	this.registry = setRegistry;

	// Images ?
	if (imgStickOn==null) {
	  imgStickOn = new ImageIcon(new ImgIcon(this,"images/StickOn.gif" ).getImage());
	  imgStickOff= new ImageIcon(new ImgIcon(this,"images/StickOff.gif").getImage());
	  imgReturn  = new ImageIcon(new ImgIcon(this,"images/Return.gif"  ).getImage());
	}

	// Begin Layout
	setLayout(new BorderLayout());
	
	// ACTION Component
	actionPanel = new JPanel();
	actionPanel.setLayout(new FlowLayout(FlowLayout.LEFT,0,0));

	actionButtonAdd    = new JButton(resources.getString("action.add"  ));
	actionButtonRemove = new JButton(resources.getString("action.del"  ));
	actionButtonUp     = new JButton(resources.getString("action.up"   ));
	actionButtonDown   = new JButton(resources.getString("action.down" ));
	actionButtonReturn = new JButton(imgReturn);

	actionCheckStick  = new JCheckBox(imgStickOff);
	actionCheckStick.setSelectedIcon (imgStickOn);
	actionCheckStick.setFocusPainted(false);
	actionCheckStick.setSelected(registry.get("sticky",false));
	
	addButtonTo(actionPanel,actionButtonAdd   ,"ADD"   ,"tip.add_prop" ,false);
	addButtonTo(actionPanel,actionButtonRemove,"DEL"   ,"tip.del_prop" ,false);
	addButtonTo(actionPanel,actionButtonUp    ,"UP"    ,"tip.up_prop"  ,false);
	addButtonTo(actionPanel,actionButtonDown  ,"DOWN"  ,"tip.down_prop",false);
	addButtonTo(actionPanel,actionButtonReturn,"RETURN","tip.return"   ,false);
	addButtonTo(actionPanel,actionCheckStick  ,"STICK" ,"tip.stick"    ,true );

	add(actionPanel,"North");
	
	// TREE Component's ScrollPane
	paneForTree = new JScrollPane();
	paneForTree.setMinimumSize(new Dimension(160, 128));
	paneForTree.setPreferredSize(new Dimension(160, 128));

	// EDIT Component			  
	panelForProxy = new JPanel();
	panelForProxy.setLayout(new BoxLayout(panelForProxy,BoxLayout.Y_AXIS));
	
	// SplitPane for top/lower section
	split = new JSplitPane(JSplitPane.VERTICAL_SPLIT,paneForTree,panelForProxy);
	split.setContinuousLayout(true);
	add(split,"Center");

	Runnable run = new Runnable() {
	  public void run() {
		int loc = registry.get("divider",-1);
		if (loc!=-1)
		  split.setDividerLocation(loc);
	  }
	};
	SwingUtilities.invokeLater(run);
	
	// Listeners
	gedcom.addListener(this);

	// Check if we can preset something to edit
	Entity lastEnt = null;
	if (!actionCheckStick.isSelected()) {
		lastEnt = gedcom.getLastEntity();
	}
	if (lastEnt==null) {
		String last = registry.get("last",(String)null);
		if (last!=null) {
			try { lastEnt = gedcom.getEntityFromId(last); } catch (Exception e) {}
		}
		if (lastEnt==null)
			try { lastEnt = gedcom.getIndi(0); } catch (Exception e) {}
		if (lastEnt==null)
			try { lastEnt = gedcom.getFam(0); } catch (Exception e) {}
	}
	if (lastEnt!=null)
		setEntity(lastEnt);
	
	// Done
}        
/**
 * One of the action buttons has been pressed
 */
public void actionPerformed(ActionEvent e) {

	// Return to last
	if ( e.getActionCommand().equals("RETURN")) {

	  // Return to last entity from return-stack
	  int last = returnStack.size()-1;
	  if (last==-1)
		return;

	  Entity old = (Entity)returnStack.elementAt(last);
	  returnStack.removeElementAt(last);
	  setEntity( old );
	  returnStack.removeElementAt(last);
	  actionButtonReturn.setEnabled(returnStack.size()>0);

	  // .. done
	  return;
	}

	// Depending on Gedcom of current entity
	if (entity==null)
	  return;
	Gedcom gedcom = entity.getGedcom();
	  
	// Add property ?	
	if (e.getActionCommand().equals("ADD")) {

	  // .. LockWrite
	  if (!gedcom.startTransaction()) {
		JOptionPane.showMessageDialog(
		  this,
		  "Couldn't save",
		  resources.getString("error"),
		  JOptionPane.ERROR_MESSAGE
		);
		return;
	  }

	  // .. Stop Editing
	  flushEditing(true);
	  
	  // .. only in case of single selection
	  TreePath paths[] = treeOfProps.getSelectionPaths();
	  if ( (paths==null) || (paths.length!=1) ) {
		gedcom.endTransaction();
		return;
	  }
	  TreePath path = treeOfProps.getSelectionPath();

	  // .. calculate new props
	  Property prop = (Property)path.getLastPathComponent();

	  // .. Confirm
	  OptionChooseProperty choose = new OptionChooseProperty(prop.getKnownProperties());
	  JCheckBox check = new JCheckBox(resources.getString("add.default_too"),true);
  
	  Object[] message = new Object[3];
	  message[0] = "Choose Property";
	  message[1] = choose;
	  message[2] = check;

	  int option = JOptionPane.showOptionDialog(this,
												message,
												resources.getString("add.title"),
												JOptionPane.OK_CANCEL_OPTION,
												JOptionPane.QUESTION_MESSAGE,
												null, null, null);
  
	  // .. OK or Cancel ?
	  if (option != JOptionPane.OK_OPTION) {
		gedcom.endTransaction();
		return;
	  }

	  // .. Calculate chosen properties
	  Property[] props = choose.getResultingProperties();

	  if ( (props==null) || (props.length==0) ) {
		JOptionPane.showMessageDialog(
		  this,
		  resources.getString("add.must_enter"),
		  resources.getString("error"),
		  JOptionPane.ERROR_MESSAGE
		);
		gedcom.endTransaction();
		return;
	  }

	  // .. add properties
	  boolean doSub = check.isSelected();
	  for (int i=0;i<props.length;i++) {
		if (doSub)
		  props[i].addDefaultProperties();
		prop.addProperty(props[i]);
	  }
  
	  // .. UnlockWrite
	  gedcom.endTransaction();

	  return;
	}

	// Remove property ?
	if (e.getActionCommand().equals("DEL")) {

	  TreePath paths[] = treeOfProps.getSelectionPaths();
	  boolean changed = false;

	  // .. check if there are some selections
	  if ( (paths==null) || (paths.length==0) ) {
		return;
	  }

	  // .. LockWrite
	  if (!gedcom.startTransaction()) {
		JOptionPane.showMessageDialog(
		  this,
		  "Couldn't save",
		  resources.getString("error"),
		  JOptionPane.ERROR_MESSAGE
		);
		return;
	  }

	  // .. Stop Editing
	  stopEditing(true);
	  
	  // .. remove every selected node      
	  for (int i=0;i<paths.length;i++) {
		
		Property prop = (Property)paths[i].getLastPathComponent();
		String veto = prop.getDeleteVeto();

		if (veto!=null) {
	  
		  JTextPane tp = new JTextPane(); 
		  tp.setText(veto);
		  tp.setEditable(false); 
		  JScrollPane sp = new JScrollPane(tp,JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,JScrollPane.HORIZONTAL_SCROLLBAR_NEVER) {
			public Dimension getPreferredSize() {
			  return new Dimension(128,64);
			}
		  };
	  
		  Object message[] = new Object[2];
		  message[0] = resources.getString("del.leads_to",prop.getTag());
	  	  message[1] = sp;
	  
		  // Show veto and respect user choice
		  int rc = JOptionPane.showConfirmDialog(
			this,
			message,
			resources.getString("warning"),
			JOptionPane.OK_CANCEL_OPTION
		  );
		  if (rc==JOptionPane.OK_OPTION)
			veto=null;

		  // Continue with/without veto
		}
		
		if (veto==null) {
		  entity.getProperty().delProperty( prop );
		  changed = true;
		}
			
		// Next selected prop
	  }

	  // .. UnlockWrite
	  gedcom.endTransaction();
	  
	  // .. done
	  return;
	}	

	// Move property UP/DOWN ? 
	if ( (e.getActionCommand().equals("UP")) || (e.getActionCommand().equals("DOWN")) ){ 

	  // .. LockWrite
	  if (!gedcom.startTransaction()) {
		JOptionPane.showMessageDialog(
		  this,
		  "Couldn't save",
		  resources.getString("error"),
		  JOptionPane.ERROR_MESSAGE
		);
		return;
	  }

	  // .. Stop Editing
	  flushEditing(true);

	  // .. Calculate property that is moved
	  entity.getProperty().moveProperty(
		currentNode,
		e.getActionCommand().equals("UP") ? Property.UP : Property.DOWN
	  );

	  // .. UnlockWrite
	  gedcom.endTransaction();

	  // 03.02.2000 Since the movement of properties is not
	  // signalled by any event, we have to reselect the node again
	  prepareTreeModel();
	  TreePath path = new TreePath(entity.getProperty().getPathTo(currentNode));
	  treeOfProps.setSelectionPath(path);

	  // .. done
	  return;
	}

	// Unknown
}
  /**
   * Helper that adds buttons to a panel
   */
  private void addButtonTo(Container c,AbstractButton b, String action, String tip, boolean enabled) {
	c.add(b);
	b.addActionListener(this);
	b.setActionCommand(action);
	b.setEnabled(enabled);
	b.setToolTipText(resources.getString(tip));
	b.setMargin(new Insets(0,0,0,0));
  }      
  /**
   * Let proxy flush editing property
   * @return status for successfull flushing of edit
   */
  boolean flushEditing(boolean alreadyLocked) {

	// Finish old editing
	if (currentProxy == null)
	  return true;

	// Prepare for finishing changed and finish
	if (currentProxy.hasChanged()) {
	  if (alreadyLocked) {
		currentProxy.finish();
	  } else {
		Gedcom gedcom = entity.getGedcom();
		if (gedcom.startTransaction()) {
		  currentProxy.finish();
		  gedcom.endTransaction();
		} else {
		  int result = JOptionPane.showConfirmDialog(
			this,
			"Couldn't save",
			resources.getString("error"),
			JOptionPane.OK_CANCEL_OPTION
		  );
		  return (result==JOptionPane.CANCEL_OPTION);
		}
	  }
	}

	// Done
	return true;
  }      
  /**
   * Notification that an existing gedcom has been closed
   */
  public void gedcomClosed(Gedcom which) {
  }      
/**
 * Notification that a change in a Gedcom-object took place.
 */
public void handleChange(Change change) {

	// Do I show an entity's properties now ?
	if (entity==null)
		return;

	// Entity deleted ?
	if ( change.isChanged(Change.EDEL) ) {

		// Loop through known entity ?
		boolean affected = false;
	  
		Enumeration ents = change.getEntities(Change.EDEL).elements();
		while (ents.hasMoreElements()) {

			Object ent = ents.nextElement();

			// ... a removed entity has to be removed from stack
			while (returnStack.removeElement(ent)) {};

			// ... and might affect the current edit view
			affected |= (ent==entity);
		}

		// Is this a show stopper at this point?
		if (affected==true) {
			setEntity(null);
			return;
		}
		
		// continue
	}
	
	// Property added/removed ?
	if ( change.isChanged(Change.PADD)
	   ||change.isChanged(Change.PDEL)) {

		prepareTreeModel();

		// .. select added
		Vector padd = change.getProperties(Change.PADD);
		if (padd.size()>0) {
			PropertyTreeModel model = (PropertyTreeModel)treeOfProps.getModel();
			Property root = (Property)model.getRoot();
			Property first = (Property)padd.firstElement();
			if (first instanceof PropertyEvent) {
				Property pdate = ((PropertyEvent)first).getDate(false);
				first = (pdate!=null) ? pdate : first;
			}
			Property[] path = root.getPathTo(first);
			if (path!=null)
				treeOfProps.setSelectionPath(new TreePath(path));
		}
		return;
	}

	// Property modified ?
	if ( change.isChanged(change.PMOD) ) {
		if ( change.getEntities(Change.EMOD).contains(entity)) {
			PropertyTreeModel treeModel = (PropertyTreeModel)treeOfProps.getModel();
			treeModel.firePropertiesChanged(change.getProperties(Change.PMOD));
			return;
		}
	}

	// Done
  }      
  /**
   * Notification that the gedcom is being closed
   */
  public void handleClose(Gedcom which) {

	// Empty stack
	returnStack.removeAllElements();

	// Leave entity
	if (entity!=null)
	  setEntity(null);

	// Forget Gedcom
	gedcom.removeListener(this);
  }      
  /**
   * Notification that an entity has been selected.
   */
  public void handleSelection(Selection selection) {

	if (actionCheckStick.isSelected())
	  return;
	setEntity(selection.getEntity());
	return;
  }      
  /**
   * Prepare the tree-model that lies behind the tree.
   */
  private void prepareTreeModel() {

	// NM 16 Dec 1998 This is some kind of hack :(
	// When a new model is set for a JTree, it signals a
	// valueChanged to its SelectionListeners, which in this
	// case means (de-)activation of buttons. That leads to
	// a focus change which leads to a redraw of the JTree.
	// When the current model has changed just a little bit
	// (which happens during GedcomWriteLock), a null-Rectangle
	// is generated somewhere in Swing which crashes the repaint 8)
	// We'll detach from JTree for the time the new model is setup
	// and attach later on

	treeOfProps.removeTreeSelectionListener(this);
	treeOfProps.setModel(new PropertyTreeModel(entity.getProperty()));
	treeOfProps.addTreeSelectionListener(this);

	// Expand all nodes
	for (int i=0;i<treeOfProps.getRowCount();i++)
	  treeOfProps.expandRow(i);

	// Done
  }      
  /**
   * Notification when component is not used any more
   */
  public void removeNotify() {

	// Remember registry
	registry.put("divider",split.getDividerLocation());
	registry.put("last", entity.getId());
	registry.put("sticky", actionCheckStick.isSelected());
	
	// Stop Listening
	gedcom.removeListener(this);

	// Continue
	super.removeNotify();
	
	// Done
  }            
  /** 
   * An entity in the gedcom data has been selected.
   * This method prepares editing of the selected entity.
   */
  void setEntity(Entity entity) {

	// Finish old editing	
	if (!stopEditing(false))
	  return;

	// Put last entity on return-stack
	if (this.entity!=null) {
	  returnStack.addElement(this.entity);
	  if (returnStack.size()>MAX_RETURN)
		returnStack.removeElementAt(0);
	  actionButtonReturn.setEnabled(returnStack.size()>0);
	}
	
	// Remember entity
	this.entity=entity; 

	// Create tree
	if (entity==null) {
	  if (treeOfProps!=null) {
		treeOfProps.removeTreeSelectionListener(this);
		treeOfProps = null;
	  }
   } else {
	  // .. create the tree
	  treeOfProps = new JTree() {
		// Calculate ToopTipText depending on property under mouse
		public String getToolTipText(MouseEvent event) {
		  // .. calc path to node under mouse
		  TreePath path = getPathForLocation(event.getX(),event.getY());
		  if ((path==null) || (path.getPathCount()==0))
			return null;
		  // .. calc property
		  Property p = (Property)path.getLastPathComponent();
		  // .. calc information text
		  String info = p.getInfo();
		  if (info==null)
			return "Unknown";
		  // .. return max 60
		  info = info.replace('\n',' ');
		  if (info.length()<=60)
			return info;
		  return info.substring(0,60)+"...";
		}
	  };
	  ToolTipManager.sharedInstance().registerComponent(treeOfProps);
	  // .. prepare data
	  prepareTreeModel();
	  // .. prepare rendering
	  treeOfProps.setCellRenderer(new PropertyCellRenderer());
	  // .. done
	}

	paneForTree.getViewport().setView(treeOfProps);
	paneForTree.validate();
	paneForTree.repaint();
	
	// Pre-selected editing node ?
	if ((entity!=null)&&(treeOfProps.isShowing()))
	  treeOfProps.setSelectionRow( 0 );

  }      
  /**
   * Prepare a proxy for editing a property
   */
  private void startEditingOf(Property prop, boolean keepSimple) {

	// New prop ?
	if (prop == null)
	  return;

	// Calculate editing for property
	String me    = getClass().getName(),
		   pkg   = me.substring( 0, me.lastIndexOf(".") + 1 ),
		   proxy = prop.getProxy();

	if (proxy == "")
	  return;

	// Create proxy    
	try {
	  if (keepSimple) throw new Exception();
	  currentProxy = (Proxy) Class.forName( pkg + "Proxy" + proxy ).newInstance();
	} catch (Exception e) {
	  currentProxy = new ProxyUnknown();
	}	

	// Add Image+Heading
	JLabel label = new JLabel();
	label.setIcon(new ImageIcon(prop.getImage(true).getImage()));
	label.setText(prop.getTag()); 
	label.setAlignmentX(0);
	label.setBorder(new EmptyBorder(2,0,8,0));
	panelForProxy.add(label); 
 
	// Add proxy components
	try {
	  currentProxy.start(panelForProxy,label,prop,this);
	} catch (ClassCastException ex) {
	  System.out.println("Wrong proxy for property "+prop);
	}

	// Layout change !
	panelForProxy.validate();
	panelForProxy.doLayout();	

	// Done
  }      
  /**
   * Stop proxy from editing property
   * @return status for successfull stop of edit
   */
  boolean stopEditing(boolean alreadyLocked) {
  
	// Finish old editing
	if (currentProxy == null)
	  return true;

	if (!flushEditing(alreadyLocked))
	  return false;

	// Clear up
	currentProxy = null;
	panelForProxy.removeAll();

	// Layout change ! 
	panelForProxy.invalidate(); 
	panelForProxy.validate(); 
	panelForProxy.repaint();	 

	// Done
	return true;
  }      
  /**
   * Called when the user changes a selection.
   * Changes the current proxy.
   */
  public void valueChanged(TreeSelectionEvent e) {

	// Look if exactly one node has been selected
	if (treeOfProps.getSelectionCount()==0) {
	  // Disable action buttons 
	  actionButtonAdd   .setEnabled(false); 
	  actionButtonRemove.setEnabled(false); 
	  actionButtonUp    .setEnabled(false); 
	  actionButtonDown  .setEnabled(false); 
	  // Done
	  return;
	}
	if (treeOfProps.getSelectionCount()>1) {
	  // En/Disable action buttons 
	  actionButtonAdd   .setEnabled(false); 
	  actionButtonRemove.setEnabled(true ); 
	  actionButtonUp    .setEnabled(false); 
	  actionButtonDown  .setEnabled(false); 
	  // Done
	  return;
	}

	
	// Calculate selection path 
	TreePath path = treeOfProps.getSelectionPath();
	
	// Prepare proxy for editing propery behind that single node      
	Property prop = (Property)path.getLastPathComponent();
	currentNode = prop;
	
	// Stop editing via old proxy before starting with new one
	if (!stopEditing(false)) {
	  // .. stop here
	  return;
	}
	startEditingOf(prop,keepSimple);

	// Enable action buttons 
	actionButtonAdd.setEnabled(true); 
	if (path.getPathCount() > 1) 
	  actionButtonRemove.setEnabled(true); 
	else
	  actionButtonRemove.setEnabled(false); 

	actionButtonUp    .setEnabled(currentNode.getPreviousSibling()!=null);
	actionButtonDown  .setEnabled(currentNode.getNextSibling()    !=null);
	
	// Done
  }      
}
