package genj.gedcom;

import java.util.*;
import java.awt.*;
import java.io.*;



import genj.util.*;/**
 * Gedcom property which constitutes a Blob
 */
public class PropertyBlob extends Property {

  private byte[]  raw;
  private String  base64;
  private ImgIcon valueAsIcon;
  private boolean isIconChecked;
  
  /**
   * Member class for iterating through adress' lines of base64-encoded data
   */
  private class Base64LineIterator implements Property.LineIterator {

	private String base64;
	private int offset;
	private final int LINE = 72;

	/**
	 * Konstruktor
	 */
	public Base64LineIterator(String base64) {
	  this.base64 = base64;
	  this.offset = 0;
	}
	
	
	/**
	 * Returns wether this iterator has more lines
	 */
	public boolean hasMoreValues() {
	  return (offset < base64.length());
	}

	/**
	 * Returns the next line of this iterator
	 */
	public String getNextValue() throws NoSuchElementException {

	  String result;
	  try {
		result = base64.substring( offset, Math.min(offset+LINE,base64.length()) );
	  } catch (StringIndexOutOfBoundsException e) {
		throw new NoSuchElementException();
	  }

	  offset+=LINE;
	  return result;
	}
	
  }
  

/**
 * Constructor of Blob Gedcom-line
 */
public PropertyBlob() {
	this("");
}  
  /**
   * Constructor of Blob Gedcom-line
   * @param in input to read data from
   */
  public PropertyBlob(File file) throws GedcomException {

	setValue(file);
	
  }      
  /**
   * Constructor of Blob Gedcom-line
   */
  public PropertyBlob(String value) {
  }      
  /**
   * Constructor of Blob Gedcom-line
   */
  public PropertyBlob(String tag, String value) {
  }      
  /**
   * Returns the data of this Blob
   */
  public byte[] getBlobData() {

	// Already present ?
	if (raw!=null)
	  return raw;

	// No Base64 present ?
	if (base64==null)
	  return null;

	// Decode Base64
	try {
	  raw = Base64.decode(base64);
	} catch (IllegalArgumentException e) {
	  return null;
	}
	
	return raw;
  }      
  /**
   * Returns the default image of this property
   */
  public static ImgIcon getDefaultImage() {
	return Images.imgBlob;
  }      
  /**
   * Returns the image of this property
   */
  public ImgIcon getImage(boolean checkValid) {
	return getDefaultImage();
  }      
  /**
   * Returns a LineIterator which can be used to iterate through
   * several lines of this address
   */
  public LineIterator getLineIterator() {

	if (raw!=null) {
	  String b64 = Base64.encode(raw);
	  return new Base64LineIterator(b64);
	}
	if (base64!=null) {
	  return new Base64LineIterator(base64);
	}

	return new Base64LineIterator("");
  }      
  /**
   * Returns the logical name of the proxy-object which knows this object
   */
  public String getProxy() {
	return "Blob";
  }      
  /**
   * Returns the name of the proxy-object which knows properties looked
   * up by TagPath
   * @return proxy's logical name
   */
  public static String getProxy(TagPath path) {
	return "Blob";
  }      
  /**
   * Returns the tag of this property
   */
  public String getTag() {
	return "BLOB";
  }      
  /**
   * Returns the property value line
   */
  public String getValue() {
	
	// Raw-Data existing ?
	if (raw!=null)
	  return raw.length+" Raw Bytes";

	// Base64-Data existing ?
	if (base64!=null)
	  return base64.length()+" Base64 Bytes";

	// None
	return "Empty";
	
  }      
  /**
   * Tries to return the data as an Icon
   */
  public synchronized ImgIcon getValueAsIcon() {

	// Already calculated?
	if (isIconChecked)
	  return valueAsIcon;
	isIconChecked = true;
	valueAsIcon   = null;
	
	// Data for Image ?
	byte[] bs = getBlobData();
	if (bs==null)
	  return null;

	// Try to create image
	Image img = Toolkit.getDefaultToolkit().createImage(bs);
	if (img!=null)
	  valueAsIcon = new ImgIcon(img);

	return valueAsIcon;
  }      
  /**
   * Yes, this property incorporates several lines as block with no newline
   */
  public int isMultiLine() {
	return MULTI_BLOCK;
  }      
  /**
   * Sets value to be taken from file
   */
  public void setValue(File file) throws GedcomException {

	// Try to open file
	FileInputStream in;
	try {
	  in = new FileInputStream(file);
	} catch (FileNotFoundException ex) {
	  throw new GedcomException("Couldn't open file "+file);
	}
	
	// Reasonable expectedSize ?
	int len = (int)file.length();

	// Read
	byte buffer[] = new byte[len];
	try {
	  len = in.read(buffer);
	} catch (IOException e) {
	  throw new GedcomException("Error while reading file "+file);
	} finally {
	  try { in.close(); } catch (Exception ie) {};
	}

	// .. nothing read ?
	if (len!=buffer.length)
	  throw new GedcomException("Couldn't read all "+buffer.length+" bytes of file "+file);

	// Successfull new information
	isIconChecked=false;
	valueAsIcon   =null;
	base64        =null;
	raw           =buffer;
	
	// Remember changed property
	noteModifiedProperty();

	// Done
  }      
  /**
   * Sets a property value line
   */
  public boolean setValue(String value) {

	// Remember value
	raw    =null;
	base64 =value.trim();
	if (base64.length()==0)
	  base64 = null;

	// Successfull new information
	isIconChecked=false;
	valueAsIcon   =null;
	
	// Remember changed property
	noteModifiedProperty();
	
	// Done
	return true;
  }      
}