package genj.gedcom;

import java.util.*; 

 
import genj.util.*;/**
  Class for encapsulating a property that describes a child-family-relationship
*/ 
public class PropertyFamilyChild extends PropertyXRef {

  /**
   * Constructor with reference
   */
  public PropertyFamilyChild(PropertyXRef target) {
	super(target);
  }      
  /**
   * Constructor with Tag,Value parameters
   */
  public PropertyFamilyChild(String tag, String value) {
	super(tag,value);
  }      
  /**
   * Returns the default image associated with this property
   */
  public static ImgIcon getDefaultImage() {
	return Images.imgLink2Fam;
  }      
  /**
   * Returns a warning string that describes what happens when this 
   * property would be deleted
   * @return warning as <code>String</code>, <code>null</code> when no warning
   */  
  public String getDeleteVeto() {
	return "The connection to the referenced family and its reference to this child are lost";
  }      
  /**
   * Returns the reference to family
   */
  public Fam getFamily() {
	return (Fam)getReferencedEntity();
  }      
  /**
   * Returns the image which is associated with this property.
   */
  public ImgIcon getImage(boolean checkValid) {
	if (checkValid&&(getReferencedEntity()==null))
	  return Images.imgLinkXFam;
	return Images.imgLink2Fam;
  }      
  /**
   * Returns the Gedcom-Tag of this property
   */
  public String getTag() {
	return "FAMC";
  }      
  /**
   * Links reference to entity (if not already done)
   * @exception GedcomException when property has no parent property,
   * referenced individual is child, wife or husband in enclosing family
   * or it would become ancestor of itself by this action.
   */
  public void link() throws GedcomException {
	
	// Something to do ?
	if (getFamily()!=null)
	  return;

	// Get enclosing individual ?
	Indi indi;
	try {
	  indi = (Indi)getEntity();
	} catch (ClassCastException ex) {
	  throw new GedcomException("FAMS can't be linked to family when not in individual");
	}

	// Prepare some VARs
	Property p;
	Property ps[];

	// Enclosing individual has a childhood already ?
	if (indi.getFamc()!=null)
	  throw new GedcomException("Individual @"+indi.getId()+"@ is already child of a family");
	
	// Look for family (not-existing -> Gedcom throws Exception)
	String id = getReferencedId();
	Fam fam = getGedcom().getFamFromId(id);
	if (fam==null)
	  throw new GedcomException("Couldn't find family with ID "+id);

	// Enclosing individual is Husband/Wife in family ?
	if ((fam.getHusband()==indi)||(fam.getWife()==indi))
	  throw new GedcomException("Individual @"+indi.getId()+"@ is already spouse in family @"+id+"@");

	// Family is descendant of indi ?
	if (fam.isDescendantOf(indi))
	  throw new GedcomException("Individual @"+indi.getId()+"@ is already ancestor of family @"+fam.getId()+"@");
	
	
	// Connect back from family (maybe using invalid back reference)
	ps = fam.getProperties(new TagPath("FAM:CHIL"),false);
	PropertyChild pc;
	for (int i=0;i<ps.length;i++) {
	  pc = (PropertyChild)ps[i];
	  if ( (!pc.isValid()) && (pc.getReferencedId().equals(indi.getId())) ) {
		pc.setTarget(this);
		setTarget(pc);
		return;
	  }
	}

	// .. new back referencing property
	pc = new PropertyChild(this);
	fam.addProperty(pc);
	setTarget(pc);
	
	// Done
  }      
  /**
   * Returns a verbose description of this property
   */
  public String toString() {
	return "Child in "+getValue();
  }      
}