package genj.gedcom;

import java.util.*;


import genj.util.*;/**
  Class for encapsulating a property that describes a spouse-family-relationship
*/ 
public class PropertyFamilySpouse extends PropertyXRef {

  /**
   * Constructor with reference
   */
  public PropertyFamilySpouse(PropertyXRef target) {
	super(target);
  }      
  /**
   * Constructor with Tag,Value parameters
   */
  public PropertyFamilySpouse(String tag, String value) {
	super(tag,value);
  }      
  public static ImgIcon getDefaultImage() {
	return Images.imgLink2Fam;
  }      
  /**
   * Returns a warning string that describes what happens when this 
   * property would be deleted
   * @return warning as <code>String</code>, <code>null</code> when no warning
   */  
  public String getDeleteVeto() {
	return "The connection to the referenced family and its reference to this spouse are lost";
  }      
  /**
   * Returns the reference to family
   */
  public Fam getFamily() {
	return (Fam)getReferencedEntity();
  }      
  /**
   * Returns the image which is associated with this property.
   */
  public ImgIcon getImage(boolean checkValid) {
	if (checkValid&&(getReferencedEntity()==null))
	  return Images.imgLinkXFam;
	return Images.imgLink2Fam;
  }      
  /**
   * Returns the Gedcom-Tag of this property
   */
  public String getTag() {
	return "FAMS";
  }      
  /**
   * Links reference to entity (if not already done)
   * @exception GedcomException when property has no parent property,
   * referenced individual is child, wife or husband in enclosing family
   * or it would become ancestor of itself by this action.
   */
  public void link() throws GedcomException {
	
	// Something to do ?
	if (getFamily()!=null)
	  return;

	// Get enclosing individual ?
	Indi indi;
	try {
	  indi = (Indi)getEntity();
	} catch (ClassCastException ex) {
	  throw new GedcomException("FAMS can't be linked to family when not in individual");
	}

	// Prepare some VARs
	Property p;
	Property ps[];
	
	// Look for family (not-existing -> Gedcom throws Exception)
	String id = getReferencedId();
	Fam fam = getGedcom().getFamFromId(id);

	if (fam==null)
	  throw new GedcomException("Couldn't find family with ID "+id);

	// Enclosing individual is Husband/Wife in family ?
	Indi husband = fam.getHusband();
	Indi wife    = fam.getWife();

	if ((husband!=null)&&(wife!=null))
	  throw new GedcomException("Family @"+fam.getId()+"@ already has husband and wife");
	
	if ((husband==indi)||(wife==indi))
	  throw new GedcomException("Individual @"+indi.getId()+"@ is already spouse in family @"+id+"@");

	if (indi.getFamc()==fam)
	  throw new GedcomException("Individual @"+indi.getId()+"@ is already child in family @"+id+"@");
	
	// Connect back from family (maybe using invalid back reference)
	PropertyHusband ph;
	ps = fam.getProperties(new TagPath("FAM:HUSB"),false);
	for (int i=0;i<ps.length;i++) {
	  ph = (PropertyHusband)ps[i];
	  if ( (!ph.isValid()) && (ph.getReferencedId().equals(indi.getId())) ) {
		if (husband!=null)
		  throw new GedcomException("Family @"+fam.getId()+"@ can't have two husbands");
		ph.setTarget(this);
		setTarget(ph);
		return;
	  }
	}
	PropertyWife pw;
	ps = fam.getProperties(new TagPath("FAM:WIFE"),false);
	for (int i=0;i<ps.length;i++) {
	  pw = (PropertyWife)ps[i];
	  if ( (!pw.isValid()) && (pw.getReferencedId().equals(indi.getId())) ) {
		if (wife!=null)
		  throw new GedcomException("Family @"+fam.getId()+"@ can't have two wifes");
		pw.setTarget(this);
		setTarget(pw);
		return;
	  }
	}

	// .. new back referencing property
	PropertyXRef px;
	if (indi.getSex()==Gedcom.MALE) {
	  if (husband!=null)
		throw new GedcomException("Family @"+fam.getId()+"@ can't have two husbands");
	  px = new PropertyHusband(this);
	  fam.addProperty(px);
	} else {
	  if (wife!=null)
		throw new GedcomException("Family @"+fam.getId()+"@ can't have two wifes");
	  px = new PropertyWife(this);
	  fam.addProperty(px);
	}
	setTarget(px);

	// Done
  }      
  /**
   * Returns a verbose description of this property
   */
  public String toString() {
	return "Spouse in "+getValue();
  }      
}