package genj.gedcom;

import java.util.*;

/**
 * Class for encapsulating a property that describes a Reference to an entity
 * @author  Nils Meier
 * @version 0.1 04/29/98
 */ 
public abstract class PropertyXRef extends Property {

  protected PropertyXRef target;
  protected String       value ;

/**
 * Constructor with reference
 * @param entity reference of entity this property links to
 */
public PropertyXRef(PropertyXRef target) {

	if (target instanceof Entity)
		throw new IllegalArgumentException("target mustn't be an Entity");
	  
	this.target = target;
	this.value  = (target==null ? "" : null);
}
  /**
   * Constructor with Tag,Value parameters
   * @param tag property's tag
   * @param value property's value
   */
  public PropertyXRef(String tag, String value) {
	setValue(value);
  }      
  /**
   * Method for notifying being removed from another parent
   */
  public void delNotify() {
  
	// Make sure the referenced entity doesn't reference back anymore
	unlink();
	
	// Let it through
	super.delNotify();
  }      
  /**
   * Returns the logical name of the proxy-object which knows this object
   * @return proxy's logical name
   */
  public String getProxy() { 
	return "XRef";
  }      
  /**
   * Returns the name of the proxy-object which knows properties looked
   * up by TagPath
   * @return proxy's logical name
   */
  public static String getProxy(TagPath path) {
	return "XRef"; 
  }      
  /**
   * Returns the entity this reference points to
   * @return entity this property links to
   */
  public Entity getReferencedEntity() {
	if (target==null)
	  return null;
	return target.getEntity();
  }      
  /**
   * Returns the id of the referenced entity
   * @return referenced entity's id
   */
  public String getReferencedId() {
	if (target==null)
	  return value == null ? "" : value;
	return target.getEntity().getId();
  }        
  /**
   * Returns the value of this property as string.
   * @return value of this property as <code>String</code>
   */
  public String getValue() { 
	if (target!=null)
	  return "@"+target.getEntity().getId()+"@";
	if ((value!=null)&&(value.length()>0))
		return "@"+value+"@";
	return "";
  }      
  /**
   * Returns <b>true</b> if this property is valid
   * @return <code>boolean</code> true or false
   */  
  public boolean isValid() {
	// Valid target?
	if (target!=null)
	  return true;
	// Empty value?
	if ((value==null)||(value.length()==0)) {
		// for example : a NOTE entity - no reference value -
		// contains data -> valid!
		return true;
	}
	return false;
  }            
  /**
   * Links reference to entity (if not already done)
   * @exception GedcomException when processing link would result in inconsistent state
   */
  public abstract void link() throws GedcomException;      
/**
 * Sets the entity's property this reference points to
 * @param target referenced entity's property
 */
protected void setTarget(PropertyXRef target) {

	if (target instanceof Entity)
		throw new IllegalArgumentException("target mustn't be of type Entity");
	  
	// Remember change
	noteModifiedProperty();
	
	// Do it
	this.target=target;
	this.value =null  ;

	// Done
}
  /**
   * Sets this property's value as string.
   */
  public boolean setValue(String value) {

	noteModifiedProperty();

	// referenced entity exists ?
	unlink();

	// remember value
	this.value=value.replace('@',' ').trim();

	// done
	return true;
  }      
  /**
   * Helper that unlinks reference to other entity.
   * removes links from other entities to this property's entity, too.
   */
  private void unlink() {

	// Referenced entities ?
	if (target==null)
	  return;

	// Forget referenced entity
	PropertyXRef t = target;
	target = null;
	
	// ... delete back referencing property in referenced entity
	//     if that property isn't the entity itself (e.g. PropertyMedia)
	if (!(t instanceof Entity)) {
	  t.getEntity().getProperty().delProperty(t);
	}
	
	// ... should be unlinked vice-versa now
  }      
}