package genj.util;

import java.io.*;
import java.net.*;
import java.util.zip.*;

/**
 * Class which stands for an origin of a resource. It is
 * basically a URL with support of a set of zipped files.
 * Examples for origins are
 *   file://home/nils/nils.ged
 *   file://home/nils/nils.zip
 */
public class Origin {

  private URL    url;

  /**
   * Sub-Class which forms a connection to an origin
   */
  public class Connection {

	private InputStream in;
	private long        len;

	Connection(InputStream in,long len) {
	  this.in=in;this.len=len;
	}

	public InputStream getInputStream() {
	  return in;
	}

	public long getLength() {
	  return len;
	}
  }
  /**
   * Constructor for given url and relative file location
   */
  public Origin(Origin o, String file) throws MalformedURLException {
	this(o.toString().substring(0, o.toString().lastIndexOf("/") + 1 ) + file.replace('\\','/'));
  }            
  /**
   * Constructor for given string location
   */
  public Origin(String s) throws MalformedURLException {
	url = new URL(s.replace('\\','/'));
  }        
  /**
   * Constructor for given url location
   */
  public Origin(URL u) throws MalformedURLException {
	this(u.toString());
  }        
/**
 * The name of this origin
 */
public String getName() {
	String u = url.toString();
	return u.substring(u.lastIndexOf('/')+1);
}      
  /**
   * Wether this origin came from file
   */
  public boolean isFile() {
	return url.toString().indexOf("file:")==0;
  }          
  /**
   * Wether this origin came from zipped file
   */
  public boolean isZip() {
	return url.getFile().endsWith(".zip");
  }      
  /**
   * Open connection to this origin
   */
  public Connection open() throws IOException {

	URLConnection uc = url.openConnection();

	return new Connection(uc.getInputStream(),uc.getContentLength());

	// Done
  }      
  /**
   * Open file relative to origin
   */
  public Connection openFile(String name) throws IOException {

	// Make sure we have '/' and not '\\'
	name = name.replace('\\','/');

	// Absolute file specification?
	if ((name.charAt(0)=='/') || (name.indexOf(":")>0) ) {

	  URLConnection uc;
	  try {
		 uc = new URL(name).openConnection();
	  } catch (MalformedURLException e1) {
		// ... hmmm, with "file://"?
		try {
		  uc = new URL("file://"+name).openConnection();
		} catch (MalformedURLException e2) {
		  return null;
		}
	  }

	  return new Connection(uc.getInputStream(),uc.getContentLength());

	}

	// File relative to origin!
	if (!isZip()) {

	  String u =  toString();
	  u = u.substring(0, u.lastIndexOf("/") +1)+name;

	  //System.out.println("Opening resource with relative path "+u);

	  URLConnection uc;
	  try {
		uc = new URL(u).openConnection();
	  } catch (MalformedURLException e) {
		throw new IOException();
	  }

	  return new Connection(uc.getInputStream(),uc.getContentLength());

	}

	// File from ZIP
	ZipInputStream in = new ZipInputStream(url.openConnection().getInputStream());
	//System.out.println("Opening resource in "+getName()+" with relative path "+name);

	// .. loop through files
	ZipEntry zentry;
	while (true) {

	  zentry = in.getNextEntry();
	  if (zentry==null)
		throw new IOException("Couldn't find resource "+name+" in ZIP-file");

	  if (zentry.getName().equals(name))
		return new Connection(in,zentry.getSize());

	}
	// Done
  }              
  /**
   * String representation
   */
  public String toString() {
	return url.toString();
  }      
/**
 * Calculates the relative path for given file if applicable
 */
public String calcRelativeLocation(String file) {

	String u = url.toString();
	int i = u.lastIndexOf('/');
	u = u.substring(0,i+1);
		
	if (!file.replace('\\','/').startsWith(u))
		return null;
	return file.substring(u.length());
}}